package io.github.townyadvanced.townyresources.controllers;

import com.gmail.goosius.siegewar.TownOccupationController;
import com.palmergames.bukkit.towny.TownySettings;
import com.palmergames.bukkit.towny.TownyUniverse;
import com.palmergames.bukkit.towny.object.Government;
import com.palmergames.bukkit.towny.object.Nation;
import com.palmergames.bukkit.towny.object.Town;
import io.github.townyadvanced.townyresources.TownyResources;
import io.github.townyadvanced.townyresources.metadata.TownyResourcesGovernmentMetaDataController;
import io.github.townyadvanced.townyresources.objects.ResourceOfferCategory;
import io.github.townyadvanced.townyresources.settings.TownyResourcesSettings;
import io.github.townyadvanced.townyresources.settings.TownyResourcesTranslation;
import io.github.townyadvanced.townyresources.util.TownyResourcesMessagingUtil;
import org.bukkit.Material;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class TownResourceProductionController {

    /**
     * Recalculate production for all towns and nations
     */
    public static void recalculateAllProduction() {
        recalculateProductionForAllTowns();
        recalculateProductionForAllNations();
        TownyResources.info("All Production Recalculated");
    }

    /**
     * Recalculate production for all towns
     * 
     * Note: This method does not recalculate production for any nations
     */
    private static void recalculateProductionForAllTowns() {
        for(Town town: TownyUniverse.getInstance().getTowns()) {           
            recalculateProductionForOneTown(town);
        }
    }

    /**
     * Recalculate production for a single town
     * 
     * Note: This method does not recalculate the nation production
     *         
     * @param town the town to recalculate production for
     */
    static void recalculateProductionForOneTown(Town town) {
        try {
            //Get discovered resources
            List<String> discoveredResources = new ArrayList<>(TownyResourcesGovernmentMetaDataController.getDiscoveredAsList(town));
    
            //Remove any discovered resources which are no longer on offer
            Map<String, ResourceOfferCategory> allOffers = TownResourceOffersController.getMaterialToResourceOfferCategoryMap();
            List<String> resourcesToRemove = new ArrayList<>();
            for(String resource: discoveredResources) {
                if(!allOffers.containsKey(resource)) {
                    resourcesToRemove.add(resource);
                }
            }
            if(!resourcesToRemove.isEmpty()) {
                discoveredResources.removeAll(resourcesToRemove);
                TownyResourcesGovernmentMetaDataController.setDiscovered(town, discoveredResources);
                town.save();
            }

            //Determine town cut
            double townCutNormalized = calculateTownCutNormalized(town);

            //Build the town production map
            Map<String, Integer> townProduction = calculateProduction(town, townCutNormalized);
    
            //Save data
            TownyResourcesGovernmentMetaDataController.setDailyProduction(town, townProduction);    
            town.save();            
        } catch (Exception e) {
            TownyResources.severe("Problem recalculating production for town" + town.getName());
            e.printStackTrace();
        }
    }

    private static double calculateTownCutNormalized(Town town) {
        if(
            town.hasNation() 
            || 
            (TownyResources.getPlugin().isSiegeWarInstalled() && TownOccupationController.isTownOccupied(town))
        ) 
            return 1 - TownyResourcesSettings.getTownResourcesProductionNationTaxNormalized();
        else
            return 1;
    }

    /**
     * Recalculate production for all nations
     * 
     * Note: This method does not recalculate production for any towns
     */
    private static void recalculateProductionForAllNations() {
        for(Nation nation: TownyUniverse.getInstance().getNations()) {
            recalculateProductionForOneNation(nation);
        }
    }

    /**
     * Recalculate production for a single nation
     *         
     * Note: This method does not recalculate production for any towns

     * @param nation the nation to recalculate production for
     */
    static void recalculateProductionForOneNation(Nation nation) {
        List<Town> townsToTakeFrom = new ArrayList<>();
        
        //Build list of towns to take from
        if(TownyResources.getPlugin().isSiegeWarInstalled()) {
            //Add all home towns, except those occupied by foreign nations
            List<Town> occupiedHomeTowns = TownOccupationController.getOccupiedHomeTowns(nation);
            for(Town town: nation.getTowns()) {
                if(!occupiedHomeTowns.contains(town) || TownOccupationController.getTownOccupier(town) == nation) {
                    townsToTakeFrom.add(town);
                }
            }
            //Add all foreign occupied towns
            List<Town> occupiedForeignTowns = TownOccupationController.getOccupiedForeignTowns(nation);               
            townsToTakeFrom.addAll(occupiedForeignTowns);
        } else {
            //Add all home towns
            townsToTakeFrom.addAll(nation.getTowns());
        }
        
        //Take resources from towns and give to nation
        double nationCutNormalized = TownyResourcesSettings.getTownResourcesProductionNationTaxNormalized();
        Map<String,Integer> nationProduction = new HashMap<>();
        Map<String,Integer> resourcesTakenFromTown;
        String takenResource;
        int takenQuantity;
        for(Town town: townsToTakeFrom) {
            //Take resources from town
            resourcesTakenFromTown = calculateProduction(town, nationCutNormalized);
            //Add resources to nation
            for(Map.Entry<String, Integer> resourceTakenFromTown: resourcesTakenFromTown.entrySet()) {
                takenResource = resourceTakenFromTown.getKey();
                takenQuantity = resourceTakenFromTown.getValue();
                if(nationProduction.containsKey(takenResource)) {
                    nationProduction.put(takenResource, nationProduction.get(takenResource) + takenQuantity);
                } else {
                    nationProduction.put(takenResource, takenQuantity);
                }
            }
        }
        
        //Set nation production & save
        TownyResourcesGovernmentMetaDataController.setDailyProduction(nation, nationProduction);
        nation.save();
    }


    /**
     * Utility Method 
     * Calculate Production
     * This can be used for a town or nation
     * 
     * @param town the town producing the resource
     * @param cutNormalized the cut of the resource to return
     * @return the production as a map, with each value multiplied by the given cutNormalized value
     */
    private static Map<String, Integer> calculateProduction(Town town, double cutNormalized) {        
        //Get all offers
        Map<String, ResourceOfferCategory> allOffers = TownResourceOffersController.getMaterialToResourceOfferCategoryMap();
        
        //Get discovered resources
        List<String> discoveredResources = new ArrayList<>(TownyResourcesGovernmentMetaDataController.getDiscoveredAsList(town));

        //Get configured resource level bonuses
        List<Double> normalizedBonusesPerResourceLevel = TownyResourcesSettings.getNormalizedProductionBonusesPerResourceLevel();

        //Calculate the production
        Map<String, Integer> production = new HashMap<>();
        String material;
        double baseProducedAmount;
        int finalProducedAmount;

        for(int i = 0; i < discoveredResources.size(); i++) {
            //Ensure town meets the town level requirement to produce the resource
            if(TownySettings.calcTownLevel(town) <  TownyResourcesSettings.getProductionTownLevelRequirementPerResourceLevel().get(i)) 
                break;
            material = discoveredResources.get(i);
            baseProducedAmount = allOffers.get(material).getBaseAmountItems();
            finalProducedAmount = (int)((baseProducedAmount * normalizedBonusesPerResourceLevel.get(i) * cutNormalized) + 0.5);
            production.put(material, finalProducedAmount);
        }
            
        return production;
    }

    /////////////////////// PRODUCING THE RESOURCES ////////////////////////////////

    /**
     * Produce all resources
     */
    public static void produceAllResources() {
        produceAllResourcesForAllTowns();
        produceAllResourcesForAllNations();
    }

    /**
     * Produce resources of all towns
     */
    private static void produceAllResourcesForAllTowns() {
        int numProducingTowns = 0;
        for(Town town: TownyUniverse.getInstance().getTowns()) {
            if(produceResourcesForOneGovernment(town))
                numProducingTowns++;
        }
        TownyResourcesMessagingUtil.sendGlobalMessage(TownyResourcesTranslation.of("production.message", numProducingTowns));        
    }

    /**
     * Produce resources of all nations
     */
    private static void produceAllResourcesForAllNations() {
        for(Nation nation: TownyUniverse.getInstance().getNations()) {
            produceResourcesForOneGovernment(nation);
        }
    }

    /**
     * Utility Method 
     * Produce resources for just one government (i.e either a town or nation)
     * 
     * @param government the government to produce resources
     * @return true if any resources were produced
     */
    private static boolean produceResourcesForOneGovernment(Government government) {
        try {
            //Get daily production
            Map<String, Integer> townDailyProduction = TownyResourcesGovernmentMetaDataController.getDailyProductionAsMap(government);
    
            if(townDailyProduction.isEmpty())
                return false;
                
            //Get the list of resources which are already available for collection
            Map<String,Integer> availableResources = TownyResourcesGovernmentMetaDataController.getAvailableForCollectionAsMap(government);
    
            //Get storage Limit modifier
            int storageLimitModifier = TownyResourcesSettings.getStorageLimitModifier();
            
            //Produce resources
            String resource;
            int quantityToProduce;
            int currentQuantity;
            int storageLimit;
            for(Map.Entry<String, Integer> townProductionEntry: townDailyProduction.entrySet()) {
                resource = townProductionEntry.getKey();
                quantityToProduce =townProductionEntry.getValue();
                if(availableResources.containsKey(resource)) {
                    //Don't go over limit
                    currentQuantity = availableResources.get(resource);
                    storageLimit = quantityToProduce * storageLimitModifier;
                    if(currentQuantity == storageLimit) {
                        continue; //Already at limit
                    } else if (currentQuantity + quantityToProduce > storageLimit) {
                        quantityToProduce = storageLimit - currentQuantity; 
                    }                       
                    //Add to existing available resources
                    availableResources.put(resource, currentQuantity + quantityToProduce);
                } else {
                    //Add new available resource
                    availableResources.put(resource, quantityToProduce);
                }
            }
    
            //Set the list of available resources
            TownyResourcesGovernmentMetaDataController.setAvailableForCollection(government, availableResources);    
            
            //Save government
            government.save();        
                           
            } catch (Exception e) {
                TownyResources.severe("Problem producing resources for government " + government.getName());
                e.printStackTrace();
                return false;
            }

        //Some resources were produced. Return true;
        return true;
    }
}
